//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Ranlux.cc:	--- implementation of Ranlux class
//
//	DESCRIPTION:
//		Implementation of Ranlux class - based on the Fortran
//		subroutine RANLUX in ranlux.f:
//		Subtract-and-borrow random number generator proposed by
//		Marsaglia and Zaman, implemented by F. James with the name
//		RCARRY in 1991, and later improved by Martin Luescher
//		in 1993 to produce "Luxury Pseudorandom Numbers".
//		Fortran 77 coded by F. James, 1993
//
//		M. Luscher, Computer Physics Communications  79 (1994) 100
//		F. James, Computer Physics Communications 79 (1994) 111
//
//	CHANGE HISTORY:
//		UL, 1999-05-27: 1.0.0: - first version
//		UL, 1999-06-08: 1.1.0: - added implementation if min()/max()
//		UL, 2002-03-02: 1.2.0: - extented class to give client control
//								 over type of range
//		UL, 2002-03-25: 1.3.0: - added minMaxRange()
//		UL, 2002-05-07: 1.3.1: - fixed bug in minMaxRange()
//		UL, 2002-11-04: 1.4.0: - Ranlux constructur honors presence of
//									SAS_RAND_SEED
//								 environment variable
//
//	AUTHOR:
//		U. Lammers (UL), Thu May 27 16:15:23 MET DST 1999
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/utils/src/Ranlux.cc,v 1.11 2002/12/16 12:03:57 jbakker Exp $
//
#include <iostream>
#include <ctime>
#include <cstdlib>

#include "Ranlux.h"
#include "cfortran.h"
#include "common.h"
#include "stringutil.h"

//
//	create wrappers for Fortran subroutines
//

//#define SLATEC_AS_MODULE  //  uncomment if SLATEC is a module

#if !defined(AS_MODULE)
	#if defined(F90MODPROC)
		#undef F90MODPROC
		#define F90MODPROC(a, b)    CXXFUNC(b, b)
	#endif
#endif



//
//	RLUX: returns a vector RVEC of LEN 32-bit random floating point numbers
//		  between zero (not included) and one (also not incl.)
//
PROTOCCALLSFSUB2(F90MODPROC(slatec, ranlux), ranlux,					\
                 FLOATV, INT)
#define RANLUX_RLUX(RVEC, LEN)											\
    CCALLSFSUB2(RANLUX, ranlux, FLOATV, INT, RVEC, LEN)

//
//  RLUXGO: initializes the generator from one 32-bit integer INT and sets
//          Luxury Level
//
PROTOCCALLSFSUB4(F90MODPROC(slatec, rluxgo), rluxgo,					\
                 INT, INT, INT, INT)
#define RANLUX_RLUXGO(LUX, SEED)                                        \
    CCALLSFSUB4(RLUXGO, rluxgo, INT, INT, INT, INT, LUX, SEED, 0, 0)


static Ranlux *ranluxInst = 0;	// single global static instance of Ranlux


void Ranlux::checkLimits() const
{
	if (_min>_max) {
		std::cerr<<"min ("<<_min<<") is larger or equal to max ("<<_max
				 <<") - cannot initialize random number generator"<<std::endl;
		exit(1);
	}
} // checkLimits()


//
//	methods
//
static const char *SAS_RAND_SEED	= "SAS_RAND_SEED";
									  // environment variable which - if
									  // defined - needs to be set to numeric
									  // values that the single global random
									  // number generator shall be instanitated
									  // with

Ranlux::Ranlux(SophisticationLevel level, RangeType rangeType,
			   unsigned seed, double min, double max)
			: RandomNumberGenerator(level, rangeType, seed, min, max),
			  _range(max-min)
{
	checkLimits();
	bool validSeed = true;
	_seed = seed;
	if (!seed) {
		validSeed = false;
		char *seedptr = getenv(SAS_RAND_SEED);
		if (seedptr)
			try {
				_seed = toNumber<unsigned>(seedptr);
				validSeed = true;
			} catch(std::string err) {
				std::cerr<<err<<" - continuing as if "
					<<SAS_RAND_SEED<<" had not been specified\n";
			}
	}

	if (!validSeed) {
		srand((unsigned)std::time((time_t *)NULL));
		_seed = std::rand();
	}

	RANLUX_RLUXGO(UNT(_level), _seed);
} // Ranlux::Ranlux()


void Ranlux::sophisticationLevel(SophisticationLevel level)
{
	_level = level;
	reset();
} // Ranlux::sophisticationLevel()


void Ranlux::rangeType(RangeType rt)
{
	_rangeType = rt;
} // Ranlux::rangeType()


void Ranlux::seed(unsigned seed)
{
	_seed = seed;
	reset();
} // Ranlux::seed()


void Ranlux::min(double min)
{
	_min = min;
	_range = _max - min;
	checkLimits();
} // Ranlux::min()


void Ranlux::max(double max)
{
	_max = max;
	_range = max - _min;
	checkLimits();
} // Ranlux::min()


void Ranlux::minMaxRange(double min, double max, RangeType rangeType)
{
	this->rangeType(rangeType);
	_min = min;
	_max = max;
	_range = max - min;
	checkLimits();
} // Ranlux::minMaxRange()


void Ranlux::reset()
{
	unsigned seed = _seed;
	SophisticationLevel level = _level;
	RangeType rangeType = _rangeType;
	double min = _min, max = _max;
	_range = max - min;
	delete ranluxInst;
	ranluxInst = new Ranlux(level, rangeType, seed, min, max);
} // Ranlux::reset()


double Ranlux::rand()
{
	const unsigned RNDNUM_BUFF_LEN = 100; // length of buffer for rnd numbers
	const double EXTRANGE = 1.e-5;

	static float rndNumBuff[RNDNUM_BUFF_LEN];
	static unsigned buffIndx = RNDNUM_BUFF_LEN;

	if (buffIndx < RNDNUM_BUFF_LEN) {
		//
		//	still unused numbers in buffer
		//
		double rnum = DBL(rndNumBuff[buffIndx++])*(1.+EXTRANGE) -
						.5*EXTRANGE,
				//	Please note: numbers on rndNumBuf are in open range (0, 1)
				//	now, -EXTRANGE/2 < rnum < 1+EXTRANGE/2
			   v = _range * rnum + _min;

		switch (_rangeType) {
			case OPEN:
				return (v>_min && v<_max ? v : rand()); break;
			case OPEN_CLOSED:
				return (v>_min && v<=_max ? v : rand()); break;
			case CLOSED_OPEN:
				return (v>=_min && v<_max ? v : rand()); break;
			case CLOSED:
				return (v>=_min && v<=_max ? v : rand()); break;
			default:
				return rand(); break;
		}
	}

	//
	//	re-fill the buffer
	//
	RANLUX_RLUX(rndNumBuff, RNDNUM_BUFF_LEN);
	buffIndx = 0;

	return rand();
} // Ranlux::rand()


RandomNumberGenerator &randGenFunc()
{
	static bool first = true;
	if (first) {
		first = false;
		ranluxInst = new Ranlux;
	}

	return *ranluxInst;
} // randGenFunc()
