//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	STime.cc	--- implementation of class STime
//
//	DESCRIPTION:
//		see STime.h
//
//	CHANGE HISTORY:
//		UL, 1997-10-23: 1.0.0: - first version
//		UL, 1998-07-13: 1.1.0: - handle leap seconds correctly; list of
//							     leap seconds should come from a data set
//							     (FTOOLS leap seconds file or CCF::LeapSeconds)
//							     reading this with the Dal would lead to
//							     a cyclic dependency Dal<->utils; to be
//							     sorted out later; as a fallback solution
//							     the list of leap seonds since 1997-07-01
//							     is hardcoded
//		UL, 1998-08-19: 1.2.0: - closed to tiny memory leaks
//							   - operator+/operator- return no longer
//								 references
//							   - bug fix in normalize()
//		UL, 1998-10-01: 1.3.0: - typo fixed: Mai -> May
//		UL, 1998-11-06: 1.4.0: - make STime constructor accept "now" as
//								 alias for STime::currentTime()
//		UL, 1999-01-18: 1.5.0: - latest leap second added
//		UL, 1999-02-02: 1.6.0: - LeapSecond object no longer STime class
//								 member; this caused initialization problems
//								 with static const STime objects in other
//								 classes
//		UL, 1999-08-06: 1.7.0: - changed STime::earliest to 1958-01-01T00:00:00
//							   - STime construction more robust now; throws
//								 exceptions upon invalid input
//		UL, 1999-08-09: 1.8.0: - added static check-function to check if
//								 STime construction will succeed
//		UL, 1999-08-10: 1.9.0: - added error handler switch: when invalid
//								 arguments are passed to STime constructors
//								 either an exception is generated or an
//								 error message is printed and execution
//								 is terminated
//							   - toString() has another optional argument
//								 determining the number of decimal digits
//								 of the fractional second part
//		UL, 1999-08-13: 1.10.0:- added alternative implementation for
//								 static leapSecond() function; original
//								 one gives core dumps (comiler bug?)
//							   - remove white-space around date/time string
//								 passed to STime constructor before it
//								 is being used (pre-1.38 behavior)
//		UL, 1999-10-21: 1.11.0:- modified again leapSecs() function to avoid
//								 crash in f90 tasks
//		UL, 2000-02-29: 1.12.0: - STime::julian/modifiedJulianDay now returns
//								 JulianDayNumber object
//		UL, 2000-06-27: 1.13.0: - workaround for Linux-egcs compiler bug
//								  in operator-(STime &, STime &)
//		UL, 2000-07-04: 1.14.0: - allow construction from Decimal Year Number
//								- various changes to fix:
//									SPR 963: Problems with STime
//									SPR 994: STime calendar format asString()
//		UL, 2000-07-05: 1.14.1: - forgotten 'ends' in formatTime() added
//								- workaround for memory leak in stringstream:
//								    char *p = s.str(); string s(p); delete p;
//		UL, 2000-11-21: 1.15.0: - fixed a few compilation problems under
//								  gcc-2.96
//		UL, 2001-01-19: 1.16.0: - STime::currentTime() == STime("now") return
//								  now GMT; they used to return local time
//		UL, 2001-06-27: 1.17.0: - fixed problem: arithmetic with times close
//								  to a minute boundary could yield a
//								  value of 60 in the sec-field, e.g.
//								  "2001-01-01:00:00:59.9999999" -> "..:60.000";
//								  (see also SPR 2081: value of 60 in second
//								  field of VALDATE)
//		UL, 2001-07-03: 1.17.1: - few tweaks to restore original accuracy
//								  of time arithmetic
//		UL, 2001-12-04: 1.18.0: - siderial->sidereal
//								- fixed problem with format of sec field
//								  (e.g. 21:47:00 was returned as "21:56:60")
//		UL, 2002-02-07: 1.18.1: - operator-(t1, t2): workaround for
//								  gcc-2.95.2/3 optimizer bug under Linux
//		UL, 2003-01-22: 1.18.2: - another gcc-optimizer-bug workaround in
//								  operator-(t1, t2); see below
//		UL, 2003-02-10: 1.19.0: - LeapSeconds::LeapSeconds is now initialized
//								  from contents of ASCII
//								  'lib/data/leapsecdata/leapsecs.dat'
//
//	AUTHOR:
//		U. Lammers, Thu Oct 23 14:15:34 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Fundamental Astronomy, H. Karttunen et. al, Springer-Verlag
//		[3]: The Astronomical Almanac for the year 2000
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/utils/src/STime.cc,v 1.38 2003/02/10 10:38:59 ulammers Exp $
///
#include <cassert>
#include <cmath>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <fstream>
#include "STime.h"
#include "stringutil.h"
#include "SasFileLocator.h"	//	for locating leap seconds file
#include "common.h"

using namespace std;

//	error handling
STime::ErrorHandling STime::errorHandling = ERRORANDEXIT;

#define STIME_ERROR(a, b)											\
{ if (STime::errorHandling == ERRORANDEXIT) {						\
	std::cerr<<"STime::error: "<<b<<" is not a valid date/time"<<std::endl;	\
	exit(1);														\
  } else { throw a(b); }											\
}

static const std::string __monthNames[] =
						{ "Jan", "Feb", "Mar", "Apr", "May",
						  "Jun", "Jul", "Aug", "Sep", "Oct",
						  "Nov", "Dec", "" };
StringIntMap<int> STime::_monthNames(__monthNames);

static const std::string __weekDayNames[] =
						{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", ""};
StringIntMap<int> STime::_weekDayNames(__weekDayNames);


// earliest and latest time
const STime STime::earliest((unsigned)1958);	// 1958-01-01T00:00:00.000
const STime STime::latest(2029, 12, 31, 23, 59, 59, 0.);
												// 2029-31-12T23:59:59.000

//
//	LeapSeconds object - the following strange construction is necessary
//	because the STime constructor - which needs a LeapSeconds obkect - is
//	called for static STime members of other classes
//	(e.g. Xmm::missionReference time); if there is a single global
//	LeapSeconds object it may not exit yet at the time the STime constructor
//	is called; had especially problem from f90 tasks
//
static LeapSeconds &leapSecs()
{
	static LeapSeconds *ls = 0;
	if (!ls)
		ls = new LeapSeconds;

	return *ls;
} // leapSecs()

//
// default constructur - initalize object to STime::earliest
//
STime::STime()
{
	*this = STime::earliest;
} // STime::STime()


//
//	construct from year/month/day/hour/minute/secs/excessSecs
//
STime::STime(unsigned year, unsigned month, unsigned day,
			 unsigned hour, unsigned minute, unsigned secs,
			 double excessSecs)
{
	if (year >= 1958 && year <= 2029)
		initMembers(year, month, day, hour, minute, secs, excessSecs);
	else {
		*this = STime::earliest;
		std::string timeString =
							::toString(year) + "-" + ::toString(month) + "-" +
							::toString(day) + "T" + ::toString(hour) + ":" +
							::toString(minute) + ":" + ::toString(secs) + "." +
							::toString(UNT(excessSecs * 1000.));
		STIME_ERROR(InvalidRange, timeString);
	}
} // STime::STime()


//
//	construct from date string "2000-30-11[T23:00:34[.234]]" or "23/12/99"
//	or calendar date format
//
STime::STime(const std::string &dateTimeString)
{
	std::string timeString(trim(dateTimeString, true));	// remove leading +
														// trailing white space
	if (lowercase(timeString) != "now") {
		*this = earliest;
		unsigned year, month, day, hour, min, secs;
		year = month = day = hour = min = secs = 0;
		double fracSecs = 0.;

		hour = min = secs = 0;
		std::string weekDay;
		std::string dateString = timeString, sep = "/";
		if (timeString.find(' ') == std::string::npos) {
			//	FITS format
			if (timeString.find('T') != std::string::npos) {
				//	"yyyy-mm-ddThh:mm:ss[.msec]"
				std::vector<std::string> v(breakup(timeString, "T"));
				if (v.size() != 2)
					STIME_ERROR(InvalidFormat, timeString);
				dateString = v[0];

				//	dissect hh:mm:ss[.msec]
				std::vector<std::string> v2(breakup(v[1], ":"));
				if (v2.size() != 3)
					STIME_ERROR(InvalidFormat, timeString);

				hour = toNumber<unsigned>(v2[0]);
				min = toNumber<unsigned>(v2[1]);
				double is;
				fracSecs = std::modf(toNumber<double>(v2[2]), &is);
				secs = UNT(is);
			}

			//	check on date string
			if (dateString.find('-') != std::string::npos)
				sep = "-";
			std::vector<std::string> v(breakup(dateString, sep));
			if (v.size() != 3)
				STIME_ERROR(InvalidFormat, timeString);

			year = toNumber<unsigned>(v[0]);
			month = toNumber<unsigned>(v[1]);
			day = toNumber<unsigned>(v[2]);
		} else {
			// CALENDAR format: "Sun Sep 16 01:03:52 1973"
			std::vector<std::string> v(breakup(timeString, " "));
			if (v.size() != 5)
				STIME_ERROR(InvalidFormat, timeString);

			weekDay = v[0];
			year = toNumber<unsigned>(v[4]);
			if (_monthNames[v[1]] >= 0)
				month = _monthNames[v[1]] + 1;
			else
				STIME_ERROR(InvalidFormat, timeString);
			day = toNumber<unsigned>(v[2]);

			//	dissect hh:mm:ss[.msec]
			std::vector<std::string> v2(breakup(v[3], ":"));
			if (v2.size() != 3)
				STIME_ERROR(InvalidFormat, timeString);

			hour = toNumber<unsigned>(v2[0]);
			min = toNumber<unsigned>(v2[1]);
			double is;
			fracSecs = std::modf(toNumber<double>(v2[2]), &is);
			secs = UNT(is);
		}

		//	range check
		if (isWithinLimits(year, month, day, hour, min, secs, fracSecs))
			initMembers(year, month, day, hour, min, secs, fracSecs);
		else
			STIME_ERROR(InvalidRange, timeString);
		//	weekday check
		if (weekDay.length() &&
			(_weekDayNames[weekDay] < 0 ||
			 _weekDayNames[weekDay] != weekday())) {
				STIME_ERROR(WrongWeekDay, timeString);
		}
	} else
		*this = currentTime();
} // STime::STime()


//
//	construct from JD/MJD number
//
static const double	julDay2modJulday		= 2400000.5;
static const unsigned julDay2modJuldayInt	= 2400000;
static const double	julDay2modJuldayFrac	= .5;

STime::STime(unsigned long timeValInt, double timeValFrac, Representation r)
{
	if (r == STime::MJD) {
		timeValInt += julDay2modJuldayInt;
		timeValFrac += julDay2modJuldayFrac;
	}

	initMembers(timeValInt, timeValFrac);
} // STime::STime()


STime::STime(double timeVal, Representation r)
{
	if (isValidTime(timeVal, r)) {

		double timeValIntD, timeValFrac = std::modf(timeVal, &timeValIntD);
		unsigned long timeValInt = static_cast<unsigned long>(timeValIntD);

		if (r == STime::MJD) {
			timeValInt += julDay2modJuldayInt;
			timeValFrac += julDay2modJuldayFrac;
		} else if (r== STime::DYN) {
			STime t(timeValInt, 1, 1);
			timeValFrac *= 365 + leapYear(timeValInt);
			timeValFrac += .5;
			timeValInt = t.julianDay().day();
		}
		initMembers(timeValInt, timeValFrac);
	} else {
		*this = STime::earliest;
		STIME_ERROR(InvalidRange, ::toString(timeVal));
	}
} // STime::STime()


//
//	destructor - nothing to do
//
STime::~STime() { }


unsigned short STime::year() const
{
	return _t.tm_year + 1900;
} // year()


unsigned short STime::month() const
{
	return _t.tm_mon + 1;	// tm_mon is in [0, 11]
} // month()


unsigned short STime::day() const
{
	return _t.tm_mday;
} // day()


unsigned short STime::hour() const
{
	return _t.tm_hour;
} // hour()


unsigned short STime::minute() const
{
	return _t.tm_min;
} // minute()


unsigned short STime::seconds() const
{
	return _t.tm_sec;
} // seconds()


unsigned short STime::weekday() const
{
	long delta = _modJuldayInt - 36204; //(=MJD(1958-01-01) which was a Wed)
	return (delta+3)%7;
} // weekday()


double STime::fracSeconds() const
{
	return _excessSecs;
} // fracSeconds()


unsigned short STime::leapSeconds() const
{
	return _nLeapSeconds;
} // leapSeconds()


const double TAI2TT = 32.184;	//	TT is that much ahead of TAI

STime::JulianDayNumber STime::julianDay(TimeSystem timeSystem) const
{
	unsigned long dayNr = _modJuldayInt + julDay2modJuldayInt;
	double fracDayNr = _modJuldayFrac + julDay2modJuldayFrac;

	if (timeSystem == TAI || timeSystem == TT)
		fracDayNr += static_cast<double>(_nLeapSeconds)/secondsPerDay;

	if (timeSystem == TT)
		fracDayNr += TAI2TT/secondsPerDay;

	return JulianDayNumber(dayNr, fracDayNr);
} // STime::julianDay()


STime::JulianDayNumber STime::modifiedJulianDay(TimeSystem timeSystem)
const
{
	unsigned long dayNr = _modJuldayInt;
	double	fracDayNr = _modJuldayFrac;

	if (timeSystem == TAI || timeSystem == TT)
		fracDayNr += static_cast<double>(_nLeapSeconds)/secondsPerDay;

	if (timeSystem == TT)
		fracDayNr += TAI2TT/secondsPerDay;

	return JulianDayNumber(dayNr, fracDayNr);
} // STime::modifiedJulianDay


double STime::decimalYearNumber() const
{
	double secsInYear = *this - STime(year(), 1, 1);
	return secsInYear/(STime::secondsPerDay*
							(daysPerNonLeapYear+leapYear(year()))) + year();
} // STime::decimalYearNumber()


double STime::hoursAsNormalizedArc() const
{
	static double TWO_PI = 2.*acos(-1.);

	double fracDay =
			(3600.*DBL(_t.tm_hour) + 60.*DBL(_t.tm_min) + DBL(_t.tm_sec) +
			_excessSecs)/86400.;

	return std::fmod(TWO_PI * fracDay, TWO_PI);
} // STime::hourAsNormalizedArc()


//
//	the following algorithm is taken from [2], p. 42
//
#if 0
double STime::siderealTime(double longitude) const
{
	static double TWO_PI = 2.*acos(-1.),
				  SIDEREAL_CLOCK_DELTA = (3.*60.+56.56)/86400. * TWO_PI;
					//	sidereal clock is 3min56.56s fast a day

	//
	//	first find the sidereal time at Greenwich 0 UT and compute
	//	the number of Julian centuries since 1900
	//
	double jd0 = INT(julianDay().full() - julDay2modJuldayFrac) +
				 julDay2modJuldayFrac,
		   T = (jd0 - 2415020.)/36525.;

	double absVal,
		   sid0UTGreen = TWO_PI*std::modf(.27691398 +
									 100.00219359 * T +
									 1.075e-6 * T * T, &absVal);

	double arcHour = hoursAsNormalizedArc();
	//
	//	sidereal clock runs 3min 56.56s fast a day - compute accumulated
	//	time difference
	//
	double deltaTheta = arcHour/TWO_PI * SIDEREAL_CLOCK_DELTA;

	//
	//	so, sidereal time in Greenwich is ...
	//
	double sidUTGreen = sid0UTGreen + arcHour + deltaTheta;

	return std::fmod(sidUTGreen + longitude, TWO_PI);
} // STime::siderealTime()
#endif


//
//	the following algorithm is taken from [3], B7
//
double STime::siderealTime(double longitude) const
{
	static double TWO_PI = 2.*acos(-1.),
				  SIDEREAL_CLOCK_DELTA = (3.*60.+56.56)/86400. * TWO_PI;
					//	sidereal clock is 3min56.56s fast a day

	double jd0 = INT(julianDay().full() - julDay2modJuldayFrac) +
				 julDay2modJuldayFrac,
		   T = (jd0 - 2451545.)/36525.,
		   sid0UTGreen = (24110.54841+T*(8640184.812866+T*(.093104-6.2e-6*T)))
				* TWO_PI/86400.;

	double arcHour = hoursAsNormalizedArc();
	//
	//	sidereal clock runs 3min 56.56s fast a day - compute accumulated
	//	time difference
	//
	double deltaTheta = arcHour/TWO_PI * SIDEREAL_CLOCK_DELTA;

	//
	//	so, sidereal time in Greenwich is ...
	//
	double sidUTGreen = sid0UTGreen + arcHour + deltaTheta,
		   sidUTGreenNorm = std::fmod(sidUTGreen + longitude, TWO_PI);

	return sidUTGreenNorm < 0. ? TWO_PI+sidUTGreenNorm : sidUTGreenNorm;
} // STime::siderealTime()


//
//	normalize all data members according to nominal bounds
//	re-calclulate MJD representation (fields: _modJuldayInt/_modJuldayFrac)
//	and leap seconds
//
void STime::normalize()
{
	double	secsAfterMinute = _t.tm_sec + _excessSecs;
	_excessSecs = 0.;
	_t.tm_sec = 0;
	double JDint, JDfrac, secsAfterMinuteInt, secsAfterMinuteFrac;
	JDint = toJulianDay(JDfrac);
	secsAfterMinuteFrac = std::modf(secsAfterMinute, &secsAfterMinuteInt);
	_nLeapSeconds = leapSecs().leapSeconds(JDint - julDay2modJulday
											 + JDfrac);
	double diffSecs = leapSecs().howManyBetween(
						JDint - julDay2modJulday + JDfrac, secsAfterMinute);

	JDfrac += (secsAfterMinuteInt - diffSecs + secsAfterMinuteFrac)/86400.;

	// normalize other fields
	STime t(static_cast<unsigned long>(JDint), JDfrac, STime::JD);

	_t = t._t;
	_excessSecs = t._excessSecs;
	_modJuldayInt = t._modJuldayInt;
	_modJuldayFrac = t._modJuldayFrac;
	_nLeapSeconds = leapSecs().leapSeconds(_modJuldayInt + _modJuldayFrac);

	//
	//	check if we are exactly on a leap second
	//
	if (diffSecs != 0. &&
		leapSecs().howManyBetween(_modJuldayInt + _modJuldayFrac, 1.) == 1) {
		_t.tm_sec = 60;
		_modJuldayFrac += 1/86400.;
	}
} // STime::normalize()


//
//	conversion from year/month/day/hour/min/secs to Julian day number
//
double STime::toJulianDay(double &JDfrac)
{
	// re-calculate JD fields
	double		day = static_cast<double>(_t.tm_mday);
	unsigned	month = _t.tm_mon + 1,
				year  = _t.tm_year + 1900;
	double f, g;

    if (month < 3) {
        f = static_cast<double>(year-1);
        g = static_cast<double>(month+12);
    } else {
        f = static_cast<double>(year);
        g = static_cast<double>(month);
    }
	double A = 2. - floor(f/100.) + floor(f/400.);
	double JDint = floor(365.25*f) + floor(30.6001*(g+1.)) + day + A +
				   1720994.;
	JDfrac = .5 + (3600.*_t.tm_hour + 60.*_t.tm_min +
				 static_cast<double>(_t.tm_sec) + _excessSecs) / 86400.;
	//	yes, in range [.5, 1.5)

	return JDint;
} // STime::toJulianDay()


//
//	initialize data members from year/month/day/hour/minute/secs tuple
//
void STime::initMembers(unsigned year, unsigned month, unsigned day,
						unsigned hour, unsigned minute, unsigned secs,
						double excessSecs)
{
	_t.tm_year	= year - 1900;
	_t.tm_mon	= month - 1;
	_t.tm_mday	= day;
	_t.tm_hour	= hour;
	_t.tm_min	= minute;
	_t.tm_sec	= secs;
	_excessSecs	= excessSecs;

	_t.tm_wday	= -1; // -1: weekday unknown
	_t.tm_yday	= -1; // -1: day in year unknown
	_t.tm_isdst	= 0;  // no Daylight Saving Time in effect

	normalize();
} // STime::initMembers()


//
//	initialize data members from JD number
//
void STime::initMembers(unsigned long JDint, double JDfrac)
{
	//	calculate MJD number and save it in data members
	assert(JDint >= julDay2modJuldayInt);
	_modJuldayInt = JDint - julDay2modJuldayInt;
	_modJuldayFrac = JDfrac - julDay2modJuldayFrac;
	double dum,
		   sec = _modJuldayFrac * secondsPerDay;
	if (std::modf(sec, &dum) > 1.-1.e-10) {
		_modJuldayFrac = (sec+1.e-10)/secondsPerDay;
		if (_modJuldayFrac >= 1.) {
			++_modJuldayInt;
			_modJuldayFrac -= 1.;
		}
	}

	// make 0 <= _modJuldayFrac < 1
	double excessDays = floor(_modJuldayFrac);
	_modJuldayFrac -= excessDays;
	_modJuldayInt += static_cast<unsigned long>(excessDays);

	unsigned long JD_int = _modJuldayInt + julDay2modJuldayInt + 1;

	double totsecs = _modJuldayFrac * secondsPerDay;
	if (1.-modf(totsecs, &dum) > 1.e-9)
		totsecs += 1.e-9;	//	round to us level

	_t.tm_hour = static_cast<int>(totsecs/secondsPerHour);
	_t.tm_min = static_cast<int>((totsecs-_t.tm_hour*secondsPerHour)/
								  secondsPerMinute);
	_excessSecs = modf(totsecs - _t.tm_hour*secondsPerHour -
					   _t.tm_min * secondsPerMinute, &dum);
	_t.tm_sec = static_cast<int>(dum);
	
	unsigned long A;
	if (JD_int < 2299161)
		A = JD_int;
	else {
		unsigned long x;

		x = (unsigned long)(((double)JD_int - 1867216.25)/36524.25);
		A = JD_int + 1 + x - x/4;
	}
	unsigned long B = A + 1524;
	unsigned long C = (unsigned long)(((double)B - 122.1)/365.25);
	unsigned long D = (unsigned long)(365.25 * (double)C);
	unsigned long E = (unsigned long)((double)(B - D)/30.6001);

	_t.tm_mday = B - D - (unsigned long)(30.6001 * E);
	_t.tm_mon = (E <= 13 ? E-1 : E-13) - 1;
	_t.tm_year = (_t.tm_mon >= 2 ? C-4716 : C-4715) - 1900;

	_t.tm_isdst = 0;	/* no daylight saving time */
	_nLeapSeconds = leapSecs().leapSeconds(_modJuldayInt + _modJuldayFrac);
} // STime::initMembers()


std::string STime::formatTime(unsigned fracSecondDigits) const
{
	std::stringstream t;
	t<<setw(2)<<setfill('0')<<_t.tm_hour<<':'
	 <<setw(2)<<_t.tm_min<<':'
	 <<setw(2)<<_t.tm_sec;

	//	handle the fractional seconds
	if (fracSecondDigits > 0) {
		t<<".";
		int n = INT(pow(10., DBL(fracSecondDigits)) + .5),
		    s = INT(DBL(n)*_excessSecs + .5);
		t<<setw(fracSecondDigits)<<(s < n ? s : n-1);
	}
	return t.str();
} // STime::formatTime()

//
//	string representation of STime object
//
std::string STime::toString(unsigned fracSecondDigits,
							StringRepresentation form) const
{
	std::stringstream t;
	switch (form) {
		case STime::FITS:
			t<<setw(4)<<setfill('0')<<_t.tm_year+1900<<'-'
			 <<setw(2)<<_t.tm_mon+1<<'-'
			 <<setw(2)<<_t.tm_mday<<'T'
			 <<formatTime(fracSecondDigits);
			break;
		case STime::CALENDAR:
			t<<_weekDayNames[weekday()]<<' ';
			t<<_monthNames[_t.tm_mon]<<' ';
			t<<_t.tm_mday<<' ';
			t<<formatTime(fracSecondDigits)<<' ';
			t<<setw(4)<<setfill('0')<<_t.tm_year+1900;
			break;
		default:
			break;
	}

	std::string timestring(t.str());

	// Uwe, is this still necessary? Jorgo
	std::string::size_type pos = timestring.find('\n');
	if (pos != std::string::npos)
		timestring.replace(pos, 1, "");	// chop-off trailing newline

	return timestring;
} // STime::toString


//
//	copy constructor
//
STime::STime(const STime &t)
{
	_t = t._t;
	_excessSecs = t._excessSecs;
	_modJuldayInt = t._modJuldayInt;
	_modJuldayFrac = t._modJuldayFrac;
	_nLeapSeconds = t._nLeapSeconds;
} // STime::STime()


//
//	assignment operator
//
STime &STime::operator=(const STime &t)
{
	if (this != &t) {
		_t = t._t;
		_excessSecs = t._excessSecs;
		_modJuldayInt = t._modJuldayInt;
		_modJuldayFrac = t._modJuldayFrac;
		_nLeapSeconds = t._nLeapSeconds;
	}

	return *this;
} // STime::operator=


//
//	add/subtract number of seconds
//
STime &STime::operator+=(double offset)
{
	_excessSecs += offset;
	normalize();

	return *this;
} // STime::operator+=


STime &STime::operator-=(double offset)
{
	return STime::operator+=(-offset);
} // STime::operator-=


//
//	the current time
//
STime STime::currentTime(bool asGMT)
{
	time_t		t = time(reinterpret_cast<time_t *>(0));
	struct tm	*tmptr;

	tmptr = asGMT ? gmtime(&t) : localtime(&t);

	return STime(tmptr->tm_year + 1900, tmptr->tm_mon + 1,
				 tmptr->tm_mday, tmptr->tm_hour, tmptr->tm_min,
				 tmptr->tm_sec);

} // STime::currentTime


//
// -------------------(	non-member operators/functions )--------------------
//

//
//	add subtract time offset in seconds
//
STime operator+(const STime &t, double offset)
{
	STime tmod(t);

	return tmod += offset;
} // operator+(const STime &, double)


STime operator-(const STime &t, double offset)
{
	return operator+(t, -offset);
} // operator-(const STime &, double)


//
//	subtract two STimes
//
double operator-(const STime &t1, const STime &t2)
{
	double t1mod = t1.modifiedJulianDay(STime::TAI).full(),
		   t2mod = t2.modifiedJulianDay(STime::TAI).full();
	if (t1mod > t2mod || t1mod == t2mod)
		//	workaround for gcc-2.92.2/3 optimizer bug under Linux:
		//	the equivalent expression: t1mod>=t2mod leads to an infinite
		//	recursion
		return STime::secondsPerDay *
			(static_cast<double>(t1._modJuldayInt - t2._modJuldayInt) +
			 t1._modJuldayFrac - t2._modJuldayFrac) +
			t1._nLeapSeconds - t2._nLeapSeconds;
	else
		//
		return -(STime::secondsPerDay *
			(static_cast<double>(t2._modJuldayInt - t1._modJuldayInt) +
			 t2._modJuldayFrac - t1._modJuldayFrac) +
			t2._nLeapSeconds - t1._nLeapSeconds);

		//	return -operator-(t2, t1);
		//	above line leads to infinte recursion with gcc-3.2.1 optimizer
		//	under Linux; the solution adopted now should be sound
}


//
//	relational/equality operators >, >=, <, <=, ==, !
//
bool operator>(const STime &t1, const STime &t2)
{
	return t1._modJuldayInt != t2._modJuldayInt
			? t1._modJuldayInt > t2._modJuldayInt
			: t1._modJuldayFrac > t2._modJuldayFrac;
} // operator>


bool operator==(const STime &t1, const STime &t2)
{
	return t1._modJuldayInt == t2._modJuldayInt &&
		   t1._modJuldayFrac == t2._modJuldayFrac;
} // operator==


bool operator!=(const STime &t1, const STime &t2)
{
	return !(t1 == t2);
} // operator!=


bool operator>=(const STime &t1, const STime &t2)
{
	return t1 > t2 || t1 == t2;
} // operator>=


bool operator<(const STime &t1, const STime &t2)
{
	return !(t1 >= t2);
} // operator<


bool operator<=(const STime &t1, const STime &t2)
{
	return t1 < t2 || t1 == t2;
} // operator<=


std::ostream &operator<<(std::ostream &os, const STime &t)
{
	return os<<t.toString();
} // operator<<



//
// ---------------------------( leap seconds )------------------------------
//
std::ostream &operator<<(std::ostream &, const LeapSeconds &);

LeapSeconds::LeapSeconds()
{
	const unsigned short TAI_UTC72 = 10;	//	number of leap seconds on
											//	01/01/72
	static double leapSecsSince72[] = {
		41317.0, 41498.0, 41683.0, 42048.0, 42413.0, 42778.0, 43144.0,
		43509.0, 43874.0, 44239.0, 44786.0, 45151.0, 45516.0, 46247.0,
		47161.0, 47892.0, 48257.0, 48804.0,
		49169.0, // 1993-07-01; TAI-UTC = 28s
	    49534.0, // 1994-07-01; TAI-UTC = 29s
	    50083.0, // 1996-01-01; TAI-UTC = 30s
	    50630.0, // 1997-07-01; TAI-UTC = 31s
		51179.0	 // 1999-01-01; TAI-UTC = 32s
	};

	string leapSecondsFile =
				fileLocator.locate("lib/data/leapsecdata/leapsecs.dat",
								   "", false, true);
	if (leapSecondsFile.length() == 0)
		//	no file found; initialize from list of know leap seconds
		for (unsigned n=0; n<NUM_OF(leapSecsSince72); ++n)
			push_back(LeapPair(leapSecsSince72[n], n+TAI_UTC72));
	else {
		ifstream lsf(leapSecondsFile.c_str());
		string line;
		unsigned n = 0,
				 lineNr = 0;
		while (getline(lsf, line) && lsf.good()) {
			lineNr++;
			line = trim(line, true);
			if (line.empty() || line[0] == '#')
				//	empty or comment line
				continue;
			double mjd;
			try {
				mjd = toNumber<double>(line);
			} catch(...) {
				cerr<<"Syntax error in leap seconds file `"
					<<leapSecondsFile<<"' line #"<<lineNr<<endl;
				exit(1);
			}
			push_back(LeapPair(mjd, TAI_UTC72 + n++));
		}
	}

	//cerr<<*this<<endl;
} // LeapSeconds::LeapSeconds()


unsigned short LeapSeconds::leapSeconds(double modifiedJulianDayNumber) const
{
	unsigned n, N=size();
	for (n=N; n>0; --n)
		if (modifiedJulianDayNumber >= (*this)[n-1].first)
			break;

	return n>0 ? (*this)[n-1].second : 0;
} // LeapSeonds::leapSeconds()


short LeapSeconds::howManyBetween(double modifiedJulianDayNumber,
								  double deltaTinSecs) const
{
	return leapSeconds(modifiedJulianDayNumber + deltaTinSecs/86400.) -
			leapSeconds(modifiedJulianDayNumber);
} // LeapSeconds::howManyBetween()

LeapSeconds::~LeapSeconds() {}

std::ostream &operator<<(std::ostream &os, const LeapSeconds &ls)
{
	os<<"LeapSeconds{\n";
	for (unsigned n=0; n<ls.size(); ++n)
		os<<ls[n].first<<' '<<ls[n].second<<endl;
	os<<'}';

	return os;
} // operator<<


bool STime::isValidTime(unsigned year, unsigned month,
						unsigned day, unsigned hour,
						unsigned minute, unsigned secs,
						double excessSecs)
{
	return STime::isWithinLimits(year, month, day, hour, minute, secs,
								 excessSecs);
} // STime::isValidTime()


bool STime::isValidTime(double timeVal, STime::Representation r)
{
	return (r != STime::JD && timeVal >= 0.) ||
		   timeVal >= julDay2modJuldayInt;
} // STime::isValid()


bool STime::isValidTime(const std::string &dateTimeString)
{
	ErrorHandling oldErr = errorHandling;
	errorHandling = EXCEPTION;
	try {
		STime s(dateTimeString);
		errorHandling = oldErr;
		return true;
	} catch (...) { }

	errorHandling = oldErr;
	return false;
} // isValidTime()


bool STime::isWithinLimits(unsigned year, unsigned month, unsigned day,
						   unsigned hour, unsigned minute, unsigned secs,
						   double excessSecs)
{
	static unsigned daysInMonth[] = { 31, 28, 31, 30, 31, 30,
									  31, 31, 30, 31, 30, 31 };

	if (year >= 1958 && year <= 2029 &&
		month >= 1 && month <= 12 &&
		day >= 1 &&
		day <= (month != 2 ? daysInMonth[month-1] : 28 + leapYear(year)) &&
		hour <= 23 &&
		minute <= 59 && secs <= 60 && excessSecs >= 0.)
		return true;

	return false;
} // isWithinLimits()


unsigned STime::leapYear(unsigned year)
{
	//	a year is a leap year if it is divisible by 4 unless it is also
	//	divisible by 100 unless it is also divisible by 400
	if (year % 4 == 0 && (year % 100 != 0 || year % 400 == 0))
		return 1;

	return 0;
} // leapYear()


void STime::JulianDayNumber::normalize()
{
	double id;

	_fractionalDay = std::modf(_fractionalDay, &id);
	_day += static_cast<unsigned long>(id);
} // JulianDayNumber::normalize()
