//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						 (c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	StringIntMap.h	--- a simple class providing associative int/string vectors
//
//	DESCRIPTION:
//		Instances of class <StringIntMap> represent integer/string maps
//		to relate strings in a vector to their respective indices and
//		vice versa. A <StringIntMap> is initialized with a vector
//		of strings (ended by "") and can then be referenced with either string
//		or integer indices using the []-notation. This yields the index
//		of the string in the original string vector or the string
//		value respectively.
//
//	USAGE EXAMPLES:
//
//	1)
//		#include <cassert>
//		#include <iostream>
//		#include <string>
//		#include "StringIntMap.h"
//
//		string Colors[] = { "green", "blue", "yellow", "" };
//		StringIntMap c(Colors);			// define the map
//
//		unsigned bluePos = c["blue"];	// get index of "blue"
//		string blue      = c[1];		// get string of index blue
//		std::cout << "blue is the " << bluePos << "th color" << c[bluePos];
//		assert(blue == c[bluePos]);
//		assert(c[c["yellow"]] == "yellow");
//
//	2)
//		#include <iostream>
//		#include "StringIntMap.h"
//
//		enum Instruments { XMM = 0, RGS1, RGS2, EMOS, EPN, OM };
//		typedef void (*MethodTable[])(int);
//		...
//		MethodTable t;
//		t[XMM] = ...
//		t[RGS1] = ...
//
//		static string _instruments[] = { "XMM", "RGS1", "RGS2", "EMOS", "EPN",
//										 "OM", "" };
//		static StringIntMap instruments(_instruments);
//
//		void f(const string &inst, const MethodTable t) {
//			int i = instruments[inst];
//
//			if (i < 0)
//				std::cout << inst << " is not a known instrument\n";
//			else {
//				std::cout << "Doing things for instrument " << instruments[i]
//					 << std::endl;
//				t(i);
//			}
//		} // f
//
//	CHANGE HISTORY:
//		UL, 1997-10-30: 1.0.0: - first version
//		UL, 1998-10-01: 1.1.0: - bug fix in constructor
//
//	AUTHOR:
//		U. Lammers, Thu Oct 30 09:15:27 WET 1997
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/utils/src/StringIntMap.h,v 1.10 2002/06/27 12:23:12 jbakker Exp $
///

#ifndef STRINGINTMAP_H
#define STRINGINTMAP_H

#include <iostream>
#include <cstring>	// strtok
#include <string>
#include <map>

#include "stringutil.h"

template<class IntegralType>
class StringIntMap {
public:
	//	default constructor
	StringIntMap() {}

	//	construct from vector of strings
	StringIntMap(const std::string[]);

	//	construct from single string - elements need to be
	//	separated by blank (" ") or newline ("\n")
	StringIntMap(const std::string &);

	~StringIntMap() {};

	IntegralType operator[](const std::string &) const;
	const std::string &operator[](IntegralType) const;
	unsigned n() const;		// number of elements in map
	std::string toString() const;

    std::ostream& write(std::ostream& os) const;
	//	friend std::ostream &operator<<(std::ostream &, const StringIntMap &);

private:
	typedef typename std::map<std::string, IntegralType, std::less<std::string > > MapType;

	MapType		 _m;
	const std::string *_str;
};

template<class IntegralType>
inline StringIntMap<IntegralType>::StringIntMap(const std::string s[])
{
	const std::string *s0 = s;
	int i = 0;
	_str = s;
	while (*s0 != "") {
		_m[*s0++] = static_cast<IntegralType>(i++);
	}
}

template<class IntegralType>
inline StringIntMap<IntegralType>::StringIntMap(const std::string &s)
{
	std::vector<std::string> vs(breakup(s, " \n\r"));

	unsigned num = vs.size();
	for (unsigned n=0; n<num; ++n)
		_m[vs[n]] = static_cast<IntegralType>(n);
}

template<class IntegralType>
inline const std::string &StringIntMap<IntegralType>::operator[]
												(IntegralType i) const
{
	return _str[i];
}

template<class IntegralType>
inline IntegralType StringIntMap<IntegralType>::operator[]
												(const std::string &s) const
{ 
	typename MapType::const_iterator i = _m.find(s);
	return i != _m.end() ? (*i).second : static_cast<IntegralType>(-1);
}

template<class IntegralType>
inline unsigned StringIntMap<IntegralType>::n() const
{
	return _m.size();
}

template<class IntegralType>
inline std::string StringIntMap<IntegralType>::toString() const
{
	std::string s("");

	for (typename MapType::const_iterator i = _m.begin(); i!=_m.end(); ++i)
		s += (*i).first + " ";

	s.replace(s.length(), 1, "");	// chop-off last blank

	return s;
}

template<class IntegralType>
inline std::ostream& StringIntMap<IntegralType>::write(std::ostream& os) const
{
	std::cout << "StringIntMap{";
	for (typename MapType::const_iterator i =
			_m.begin(); i != _m.end(); ++i)
		std::cout<<"(" << (*i).first << ", " << (*i).second << ")\n";
	std::cout<<"}\n";

	return os;
}

template<class IntegralType>
std::ostream& operator<<(std::ostream& os, const StringIntMap<IntegralType> &m)
{ return m.write(os); } 

#endif

